package jp.muziyoshiz.wmi.web;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jp.muziyoshiz.wmi.web.jdo.Information;
import jp.muziyoshiz.wmi.web.jdo.InformationIdCounter;
import jp.muziyoshiz.wmi.web.jdo.InformationPositionCounter;
import jp.muziyoshiz.wmi.web.jdo.Knower;
import jp.muziyoshiz.wmi.web.jdo.PMF;

import com.google.appengine.api.users.User;
import com.google.appengine.api.users.UserService;
import com.google.appengine.api.users.UserServiceFactory;

@SuppressWarnings("serial")
public class InformationServlet extends HttpServlet {
    public void doGet(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        // テスト用に何のページなのか表示
        resp.getWriter().println("<p>Information</p>");

        // ユーザ認証必須にする
        UserService userService = UserServiceFactory.getUserService();
        String thisURL = req.getRequestURI();
        if (req.getUserPrincipal() == null) {
            resp.getWriter().println(
                    "<p>Please <a href=\"" + userService.createLoginURL(thisURL)
                            + "\">sign in</a>.</p>");
            return;
        } else {
            resp.getWriter().println(
                    "<p><a href=\"" + userService.createLogoutURL(thisURL) +
                    "\">logout</a></p>");
        }

        // ユーザIDその他を取得
        User currentUser = userService.getCurrentUser();

        // テスト用のユーザIDを表示
        resp.getWriter().println("<p>UserId: " + currentUser.getUserId() + "</p>");

        // 全取得
        Collection<Information> results = getInformationByUser(currentUser);
            
        PrintWriter writer = resp.getWriter();
        writer.println("<table border=\"1\">");
        writer.println("<tr><th>Key</th><th>ID</th><th>Name</th><th>Value</th><th>Position</th><th>Knower(relation)</th></tr>");
        
        for (Information info : results) {
            writer.println("<tr>");
            writer.println("<td>" + info.getKey() + "</td>");
            writer.println("<td>" + info.getId() + "</td>");
            writer.println("<td>" + info.getName() + "</td>");
            writer.println("<td>" + info.getValue() + "</td>");
            writer.println("<td>" + info.getPosition() + "</td>");
            writer.println("<td>");
            for (Knower knower : getKnowersByInformation(info)) {
                writer.println(knower.getName());
                writer.println(", ");
            }
            writer.println("</td>");
            writer.println("</tr>");
        }
        
        writer.println("</table>");
        
        resp.getWriter().println("<a href=\"/information.jsp\">Add information</a>");
        resp.getWriter().println("<a href=\"/\">Back to index</a>");
    }

    public void doPost(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        // ユーザ認証必須にする
        UserService userService = UserServiceFactory.getUserService();
        if (req.getUserPrincipal() == null) {
            String thisURL = req.getRequestURI();
            resp.getWriter().println(
                    "<p>Please <a href=\"" + userService.createLoginURL(thisURL)
                            + "\">sign in</a>.</p>");
            return;
        }

        // ユーザIDその他を取得
        User currentUser = userService.getCurrentUser();

        String name = null;
        String value = null;

        if (req.getParameter("name") != null) {
            name = req.getParameter("name");
        }
        if (req.getParameter("value") != null) {
            value = req.getParameter("value");
        }

        // idを割り当て
        Long id = getNewInformationIdOfUser(currentUser);
        
        // informationを作成
        Information info = new Information();
        info.setKey(currentUser.getUserId() + "/" + id);
        info.setId(id);
        info.setName(name);
        info.setValue(value);
        info.setPosition(getNewInformationPositionOfUser(currentUser));

        PersistenceManager pm = PMF.createManager();
        try {
            pm.makePersistent(info);
        } finally {
            pm.close();
        }

        resp.sendRedirect("/information");
    }

    @SuppressWarnings("unchecked")
    private Collection<Knower> getKnowersByInformation(Information info) {
        // 全取得
        PersistenceManager pm = PMF.createManager();

        try {
            // 検索対象はInformation
            Query query = pm.newQuery(Knower.class);
            
            // キーの前方一致
            query.setFilter("informationKeys == :infoKey");
            
            try {
                Collection<Knower> results =
                    (Collection<Knower>)query.execute(info.getKey());
                
                return pm.detachCopyAll(results);
            } finally {
                query.closeAll();
            }
        } finally {
            pm.close();
        }
    }
    
    @SuppressWarnings("unchecked")
    public static Collection<Information> getInformationByUser(User user) {
        // 全取得
        PersistenceManager pm = PMF.createManager();

        try {
            // 検索対象はInformation
            Query query = pm.newQuery(Information.class);
            
            // キーの前方一致
            query.setFilter("key >= :keyStart && key < :keyEnd");
            
            // positionの小さい順
            // こうソートしたいけど、GAEの制約で無理（keyで部分一致させているので、
            // key以外の列ではソートできない）
//            query.setOrdering("position asc");
            
            try {
                // 検索のプレフィックス
                String prefix = user.getUserId() + "/";
                
                Collection<Information> results =
                    (Collection<Information>)query.execute(
                            prefix, prefix + '\uFFFD');
                
                return pm.detachCopyAll(results);
            } finally {
                query.closeAll();
            }
        } finally {
            pm.close();
        }
    }
    
    private Long getNewInformationIdOfUser(User user) {
        PersistenceManager pm = PMF.createManager();
        try {
            pm.currentTransaction().begin();
            try {
                InformationIdCounter counter = null;
                try {
                    counter = pm.getObjectById(InformationIdCounter.class, user.getUserId());
                } catch (JDOObjectNotFoundException e) {
                    // 見つからなかった場合は新たに定義
                    counter = new InformationIdCounter(user.getUserId(), 0L);
                }
                
                counter.increment();
                pm.makePersistent(counter);
                pm.currentTransaction().commit();
                
                return counter.getCounter();
                
            } finally {
                if (pm.currentTransaction().isActive()) {
                    pm.currentTransaction().rollback();
                }
            }
        } finally {
            pm.close();
        }
    }
    
    private Long getNewInformationPositionOfUser(User user) {
        PersistenceManager pm = PMF.createManager();
        try {
            pm.currentTransaction().begin();
            try {
                InformationPositionCounter counter = null;
                try {
                    counter = pm.getObjectById(InformationPositionCounter.class, user.getUserId());
                } catch (JDOObjectNotFoundException e) {
                    // 見つからなかった場合は新たに定義
                    counter = new InformationPositionCounter(user.getUserId(), 0L);
                }
                
                counter.increment();
                pm.makePersistent(counter);
                pm.currentTransaction().commit();
                
                return counter.getCounter();
                
            } finally {
                if (pm.currentTransaction().isActive()) {
                    pm.currentTransaction().rollback();
                }
            }
        } finally {
            pm.close();
        }
    }
}
